-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
-----
-- by @MegaKnight and @vouthin
--
-- lua tornadoes based on wind and particle emitters
--
-- DO NOT REUSE CODE WITHOUT PERMISSION
-----

local M = {}


-----
-- VARIABLES
-----

local nadoActive = false --> is active ?

local nadoCoords = {} --> coordinates of tornado, z all the way
nadoCoords.x = nil
nadoCoords.y = nil

local nadoDuration = nil --> duration of tornado in seconds


-- dust particle emitter
local dustEmitter = nil

-- forest wind emitter --> was experimental, looks weird, uncomment if you like ;)
--local forestWind = nil

-- move tornado position every 2 seconds
local moveTimer = 0


-- tornado following player or random destination ?
local pursueRandomPath = false
local destination = Point3F(0, 0, 0)

-- check if random destination reached
local oldXdestVal = 0
local newXdestVal = 0
local oldYdestVal = 0
local newYdestVal = 0
local firstMove = true

-- speed of movement
local nadoSpeed = 2 --> standard value

-- debug drawer
local isDebugEnabled = false
local debugMode = 0 --> 0 = rotational force, 1 = updraft force, 2 = direction of rotation


-- tornado parameters, by @vouthin
local tornadoRadius = 10 -- how big the tornado (in this case, the funnel) is, outer windfield is based on this value --> 10
local windSpeed = 640 -- self explanatory (mph) --> 640
local updraftSpeed = 110 -- updraft speed (I THINK mph) --> 110

local inflowMultiplier = -0.6 -- a multiplier for how strong the "pull" to the tornado is --> -0.6

local scalingFactor = 8 -- change this if youd like a smaller outer windfield --> 8
local nadoRadius = tornadoRadius * scalingFactor

-- directon of rotation
local reverseRotation = false


-- set values, for UI control
local isSetStrength = false --> should tornado strength be random or set ?
local givenRadius = 10 --> standard
local givenSpeed = 640
local givenUpdraft = 110
local givenInflow = -0.6
local givenScaling = 8
local givenRotation = false



-- set random starting values
local function setRandomValues()
	tornadoRadius = math.random(5,15)
	windSpeed = tornadoRadius * 64 * math.log10(100/tornadoRadius)
	updraftSpeed = tornadoRadius * 11
	scalingFactor = math.random(6,10)
	nadoRadius = tornadoRadius * scalingFactor
	inflowMultiplier = -0.6
	-- rotation
	local rot = math.random(1,2)
	if rot == 1 then reverseRotation = false else reverseRotation = true end
end


-- set given values from UI input
local function setGivenValues()
	tornadoRadius = givenRadius
	windSpeed = givenSpeed
	updraftSpeed = givenUpdraft
	inflowMultiplier = givenInflow
	scalingFactor = givenScaling
	nadoRadius = tornadoRadius * scalingFactor
	reverseRotation = givenRotation
end


-----
-- CORE TORNADO LOGIC
-----

-- initialize tornado
local function startTornado()

	-- don't do anything when already active tornado
	if nadoActive then return end
	
	-- set coordinates, random offset from player location
	local playerCoords = be:getPlayerVehicle(0):getPosition()
	nadoCoords.x = playerCoords.x + math.random(-100,100)
	nadoCoords.y = playerCoords.y + math.random(-100,100)
	
	-- get terrain height at (x,y), asssume terrain above z = 2000
	local zCoord = be:getSurfaceHeightBelow(vec3(nadoCoords.x, nadoCoords.y, 2000))
	
	-- set duration in seconds
	nadoDuration = math.random(120,900) --> random duration between 2 and 15 minutes
	
	-- set random destination, if random paths enabled
	if pursueRandomPath then
		destination = Point3F(playerCoords.x + math.random(-200,200), playerCoords.y + math.random(-200,200), 0)
		
		-- reset variables
		local oldXdestVal = 0
		local newXdestVal = 0
		local oldYdestVal = 0
		local newYdestVal = 0
		local firstMove = true
	end
	
	-- reset variables
	moveTimer = 0
	
	
	-- set rotation
	
	-- set values
	if isSetStrength then
		setGivenValues() --> use values from UI
	else
		setRandomValues()
	end
	
	-- activate
	nadoActive = true
	
	-- place and enable particles
	dustEmitter:setPosition(vec3(nadoCoords.x, nadoCoords.y, zCoord - 2))
	dustEmitter.emitter = scenetree.findObject("BNGP_tornado_mk") --> is nil first time ???
	dustEmitter.active = true
	scenetree.findObjectById(dustEmitter:getId()):postApply()
	
	-- place and enable forest wind
	--forestWind:setPosition(vec3(nadoCoords.x, nadoCoords.y, zCoord - 2))
	--forestWind.radius = tornadoRadius * 4
	--forestWind.strength = windSpeed / 10
	--forestWind.windEnabled = true
	--scenetree.findObjectById(forestWind:getId()):postApply()
	
	-- log nado pos, duration, radius and strength
	log("I", "", "tornado spawned: pos(" .. string.format("%.2f",nadoCoords.x) .. "," .. string.format("%.2f",nadoCoords.y) .. "), duration = " .. nadoDuration .. ", radius = " .. string.format("%.2f",tornadoRadius) .. ", strength = " .. windSpeed)
	
	-- display message
	guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "A tornado is forming! Strength: ~" .. string.format("%.0f",(windSpeed/3)) .. " mph", config = {timeOut = 5000}})
	
end


-- lil rankine vortex formula, by @vouthin
local function getTornadoFlow(dx, dy, torRadius, outerRadius, maxTangential, maxUpdraft, inflowFactor)
    local r = math.sqrt(dx^2 + dy^2)
    
    -- getting the angle relative to the center of the tornado coordinates, so that we then calculate windspeed
    local angle = math.atan2(dy, dx)
    local Vtheta = 0

    -- making the core have weaker winds cuz duhhh vortex,,,,,,,,,,,
    if r <= torRadius then
        Vtheta = maxTangential * (r / torRadius)
    else
        Vtheta = maxTangential * (torRadius / r)
    end
    if r > outerRadius then
        Vtheta = Vtheta * math.exp(-2 * (r - outerRadius) / outerRadius)
    end

    -- horizontal winds around the tornado
    local swirlVX = -Vtheta * math.sin(angle)
    local swirlVY =  Vtheta * math.cos(angle)
	
	-- reverse direction of rotation
	if reverseRotation then 
		swirlVX = - swirlVX
		swirlVY = - swirlVY
	end

    -- inflow, pulls the vehicle INTO the tornado
    local inflow = inflowFactor * Vtheta
    local inflowVX = inflow * math.cos(angle)
    local inflowVY = inflow * math.sin(angle)

    -- adding up so they both affect the car
    local vx = swirlVX + inflowVX
    local vy = swirlVY + inflowVY

    -- the updraft is strongest at a narrow area just outside the funnel, thus why im getting a "funnel" radius
    local peakRadius = torRadius * 1.5  -- peak slightly outside the funnel

    local vz = maxUpdraft * math.exp(-0.01 * (r - peakRadius)^2)
    
    return vx, vy, vz
end


-- tornado winds applier
local function doTornadoStep()
	
	-- check for vehicles in domain of influence and apply force
	for i = 0, (be:getObjectCount() - 1) do 
		local veh = be:getObject(i)
		if veh then --> does veh exist?
			local vehPos = veh:getPosition()	
			
			-- TODO:
			-- prepare for customness
			-- custom tornado from UI
			
			-- if above tornado, return
			if nadoCoords.z then
				local distZ = nadoCoords.z + 50 - vehPos.z
				if distZ < 0 then
					-- remove wind
					dynamicWeatherMK_common.applyWind(veh, Point3F(0, 0, 0))
					return
				end
			end
			
            local vehPos = veh:getPosition()
            local dx = vehPos.x - nadoCoords.x
            local dy = vehPos.y - nadoCoords.y --> calculate distance
            local vx, vy, vz = getTornadoFlow(dx, dy, tornadoRadius, nadoRadius, windSpeed, updraftSpeed, inflowMultiplier)
            dynamicWeatherMK_common.applyWind(veh, Point3F(vx, vy, vz))
			
			-- old (bad) logic (not the first, this was never in public version)
			-- https://github.com/Canadian-Severe-Storms-Laboratory/Rankine_Tornado_Vortex_Solver/blob/main/Rankine%20Vortex%20Solver%20Math.pdf
		end
	end
	
	
end


-- debug drawer logic
local sphereSize = 1
local sphereSpacing = 4
local cacheOldMode = nil
local wasMap = false
local wasActive = false
local function onPreRender()
	if not nadoActive then return end

	-- show nado field if bigmap
	if freeroam_bigMapMode then
		if freeroam_bigMapMode.bigMapActive() and (not wasMap) then 
			if isDebugEnabled then wasActive = true end
			isDebugEnabled = true 
			sphereSize = 2
			sphereSpacing = 4
			cacheOldMode = debugMode
			debugMode = 0
			wasMap = true
		elseif (not freeroam_bigMapMode.bigMapActive()) and wasMap then 
			wasMap = false
			if not wasActive then isDebugEnabled = false end
			wasActive = false
			sphereSize = 1
			sphereSpacing = 4
			debugMode = cacheOldMode
		end
	end
	
	-- get z-coordinate
	local zHeight = 50
	if nadoCoords.z then zHeight = nadoCoords.z + 50 end
	
	if not isDebugEnabled then return end
	-- visualize destination
	if pursueRandomPath then
		debugDrawer:drawSphere(Point3F(destination.x, destination.y, be:getSurfaceHeightBelow(vec3(destination.x, destination.y, 2000))), 10, ColorF(1,1,1,1))
	end
	-- iterate over coordinates
	local i = -200
	while i < 201 do
		local j = -200
		while j < 201 do
			local pX = nadoCoords.x + i
			local pY = nadoCoords.y + j
			-- calc distance and windforce
			local dist = math.sqrt((i * i) + (j * j))
			local force = Point3F(0, 0, 0)
			local vx, vy, vz = getTornadoFlow(i, j, tornadoRadius, nadoRadius, windSpeed, updraftSpeed, inflowMultiplier)
			
			-- DEBUG MODES --
			
			-- draw debug sphere based on rotational force
			if debugMode == 0 then
			if math.abs(vx) > 200 or math.abs(vy) > 200 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,1,1))
			elseif math.abs(vx) > 120 or math.abs(vy) > 120 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,0,1))
			elseif math.abs(vx) > 60 or math.abs(vy) > 60 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,1,0,1))
			elseif math.abs(vx) > 20 or math.abs(vy) > 20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,1,0,1))
			else
				if not wasMap then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,1,1,1))
				end
			end
			
			-- draw debug sphere based on upwards force
			elseif debugMode == 1 then
			if math.abs(vz) > 110 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,1,1))
			elseif math.abs(vz) > 80 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,0,1))
			elseif math.abs(vz) > 50 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,1,0,1))
			elseif math.abs(vz) > 20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,1,0,1))
			else
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,1,1,1))
			end
			
			-- draw debug sphere based on direction of rotational force
			elseif debugMode == 2 then
			if vx < -120 then 
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,0,1,1))
			elseif vx < -60 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0.5,0,1,1))
			elseif vx < -20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,1,1))
			elseif vx == 0 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(0,1,0,1))
			elseif vx < 20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,0.5,1))
			elseif vx < 60 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0,0,1))
			elseif vx < 120 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,0.5,0,1))
			else
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight), sphereSize, ColorF(1,1,0,1))
			end
			if vy < -120 then 
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(0,0,1,1))
			elseif vy < -60 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(0.5,0,1,1))
			elseif vy < -20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(1,0,1,1))
			elseif vy == 0 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(0,1,0,1))
			elseif vy < 20 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(1,0,0.5,1))
			elseif vy < 60 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(1,0,0,1))
			elseif vy < 120 then
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(1,0.5,0,1))
			else
				debugDrawer:drawSphere(Point3F(pX, pY, zHeight + 50), sphereSize, ColorF(1,1,0,1))
			end
			end
			
			j = j + sphereSpacing
		end
		i = i + sphereSpacing
	end	
end


-- moves tornado
local function moveTornado()
	
	if pursueRandomPath then -- move to random destination
		-- store old vals
		oldXdestVal = newXdestVal
		oldYdestVal = newYdestVal
		
		if nadoCoords.x > destination.x then
			nadoCoords.x = nadoCoords.x - 2
			newXdestVal = 1
		else
			nadoCoords.x = nadoCoords.x + 2
			newXdestVal = 2
		end
		if nadoCoords.y > destination.y then
			nadoCoords.y = nadoCoords.y - 2
			newYdestVal = 1
		else
			nadoCoords.y = nadoCoords.y + 2
			newYdestVal = 2
		end
		-- store new vals if first move
		if firstMove then
			oldXdestVal = newXdestVal
			oldYdestVal = newYdestVal
			firstMove = false
		end
		-- check if destination reached
		if oldXdestVal == newXdestVal and oldYdestVal == newYdestVal then
			-- nothing, destination not reached
		else
			-- set new destination
			destination = Point3F(destination.x + math.random(-100,100), destination.y + math.random(-100,100), 0)
		end
	else
		-- get player coordinates
		local playerCoords = getPlayerVehicle(0):getPosition()
		-- move tornado towards player
		if nadoCoords.x > playerCoords.x then
			nadoCoords.x = nadoCoords.x - 2
		else
			nadoCoords.x = nadoCoords.x + 2
		end
		if nadoCoords.y > playerCoords.y then
			nadoCoords.y = nadoCoords.y - 2
		else
			nadoCoords.y = nadoCoords.y + 2
		end
	end
	
	-- get terrain height at (x,y), asssume terrain above z = 2000
	nadoCoords.z = be:getSurfaceHeightBelow(vec3(nadoCoords.x, nadoCoords.y, 2000))
	
	-- move particle emitter
	dustEmitter:setPosition(vec3(nadoCoords.x, nadoCoords.y, nadoCoords.z - 1))
	scenetree.findObjectById(dustEmitter:getId()):postApply()
	
	-- move forest wind emitter
	--forestWind:setPosition(vec3(nadoCoords.x, nadoCoords.y, nadoCoords.z - 1))
	--scenetree.findObjectById(forestWind:getId()):postApply()
end


-- removes tornado
local function removeTornado()
	-- setwind force to 0
	log("I", "", "removed tornado.")
	dynamicWeatherMK_common.applyWindAll( Point3F(0, 0, 0))
	
	nadoActive = false
	
	-- hide particles
	dustEmitter.active = false
	scenetree.findObjectById(dustEmitter:getId()):postApply()
	
	-- stop wind
	--forestWind.windEnabled = false
	--scenetree.findObjectById(forestWind:getId()):postApply()
end


-- creates tornadoes randomly
local function randomTornado()
	local randomVal = math.random()
	log("I", "", "Rolled random tornado with value " .. randomVal)
	if randomVal > 0.85 then
		
		-- start tornado
		startTornado()
	end
end


-- executed each GFX frame
local doRandomTornadoes = false --> do random tornadoes ?
local randomTornadoTimer = 0
local function onUpdate(dt)

	if not simTimeAuthority.getPause() then
	local timerFactor = simTimeAuthority.getReal()

	-- do tornado logic only when tornado is active
	if nadoActive then
	
		--debugDrawer:drawSphere(Point3F(nadoCoords.x, nadoCoords.y, nadoCoords.z), 10, ColorF(1,0,0,1))
	
		--nadoTimer = nadoTimer + (dt * timerFactor)
		moveTimer = moveTimer + (dt * timerFactor)
		nadoDuration = nadoDuration - (dt * timerFactor)
		
		--if nadoTimer > 0.1 then --> execute tornado logic each 0.1 seconds --> not needed, is weird
		doTornadoStep() -- TODO: only when game unpaused!!!
			--nadoTimer = 0
		--end
		
		if moveTimer > nadoSpeed then
			moveTimer = 0
			moveTornado()
		end
		
		if nadoDuration < 0 then
			nadoActive = false --> if timer runs out, deactivate and remove tornado
			removeTornado()
		end
	
	end
	
	-- random tornado, for weather
	if doRandomTornadoes and (not nadoActive) then 
		randomTornadoTimer = randomTornadoTimer + (1 * timerFactor)
		if randomTornadoTimer > 200 then
			randomTornado() 
			randomTornadoTimer = 0
		end
	end
	
	end
end


-----
-- LOAD/UNLOAD EXTENSION
-----

local function onExtensionLoaded()
	log("I", "", "Tornado extension loaded")
	
	-- does particle emitter exist?
	dustEmitter = scenetree.findObject("tornado_dust_MKDW")
	
	if not dustEmitter then
		-- if not, create particle emitter
		dustEmitter = createObject("ParticleEmitterNode")
		dustEmitter.scale = Point3F(1, 1, 1)
		--dustEmitter.position = Point3F(1, 1, 1) --> error, seems to be matrix instead of Point3F/table
		dustEmitter.emitter = scenetree.findObject("BNGP_tornado_mk") --> custom particle
		dustEmitter.dataBlock = scenetree.findObject("lightExampleEmitterNodeData1")
		dustEmitter.velocity = 1.0
		dustEmitter.active = false
		dustEmitter.hidden = false
		dustEmitter:registerObject('tornado_dust_MKDW')
	end
	
	-- create local forest wind
	--if not forestWind then
	--	forestWind = createObject("ForestWindEmitter")
	--	forestWind.windEnabled = false
	--	forestWind.radialEmitter = true
	--	forestWind.strength = 10.0
	--	forestWind:registerObject('ForestTornadoEmitter_MKDW')	
	--end
	
end

local function onExtensionUnloaded()
	log("I", "", "Tornado extension unloaded")
end


-- called by dynamic weather, prevent crashes
local function cleanDisable()
	-- remove wind
	be:queueAllObjectLua('obj:setWind(0, 0, 0)')
	
	-- delete particles
	if scenetree.findObjectById(dustEmitter:getId()) then
		scenetree.findObjectById(dustEmitter:getId()):delete()
	end
	
	--delete wind
	--if scenetree.findObjectById(forestWind:getId()) then
	--	scenetree.findObjectById(forestWind:getId()):delete()
	--end
	
	-- unload
	extensions.unload('dynamicWeatherMK_tornado')
end


-----
-- FOR USER CONFIGURATION
-----

-- toggle random path / follow player
local function toggleRandomPath()
	if nadoActive then return end --> dont interfere with active tornado
	pursueRandomPath = not pursueRandomPath
	if pursueRandomPath then
		guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Tornado moving randomly", config = {timeOut = 5000}})
	else
		guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Tornado following player", config = {timeOut = 5000}})
	end
end


-- sets tornado speed
local function setTornadoMovementSpeed(value)
	if value > 0 and value < 11 then
		nadoSpeed = value
	else
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Please use a value between 1 and 10", config = {timeOut = 5000}})
	end
end


-- sets debug mode
local function setDebugMode(mode)
	if mode >= 0 and mode < 3 then
		debugMode = mode
		if mode == 0 then return "Tangential Force"
		elseif mode == 1 then return "Upward Force"
		else return "Direction of Rotation"
		end
	else
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Not available. Please use 0, 1 or 2", config = {timeOut = 5000}})
		return nil
	end
end


-- set tornado parameters
local function setParameters(pTable)
	-- tornadoRadius
	if pTable.radius >= 1 and pTable.radius <= 50 then
		givenRadius = pTable.radius
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Radius should be between 1 and 50", config = {timeOut = 5000}})
	end
	-- windspeed
	if pTable.speed >= 0 and pTable.speed <= 1000 then
		givenSpeed = pTable.speed
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Windspeed should be between 0 and 333", config = {timeOut = 5000}})
	end
	-- updraft speed
	if pTable.updraft >= 0 and pTable.updraft <= 1000 then
		givenUpdraft = pTable.updraft
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Updraft windspeed should be between 0 and 333", config = {timeOut = 5000}})
	end
	-- inflow multiplier
	if pTable.inflow >= -30 and pTable.inflow <= 30 then
		givenInflow = pTable.inflow
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Inflow Multiplier should be between -30 and 30", config = {timeOut = 5000}})
	end
	-- scaling factor
	if pTable.scaling >= 0 and pTable.scaling <= 30 then
		givenScaling = pTable.scaling
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Scaling Factor should be between 0 and 30", config = {timeOut = 5000}})
	end
	-- rotation
	if pTable.rotation == true or pTable.rotation == false then
		givenRotation = pTable.rotation
	else 
		guihooks.trigger('toastrMsg', {type = "error", title = "Dynamic Weather", msg = "Direction of Rotation should be true or false", config = {timeOut = 5000}})
	end
		
end



-- public interface

M.onUpdate = onUpdate --> executed by game engine
M.onPreRender = onPreRender --> debug drawer
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
--M.onClientEndMission = onClientEndMission
M.cleanDisable = cleanDisable

M.toggleDebug = function() isDebugEnabled = not isDebugEnabled end
M.isDebugEnabled = isDebugEnabled
M.setDebugMode = setDebugMode

M.startTornado = startTornado --> console interaction
M.removeTornado = removeTornado
M.isNadoActive = function() return nadoActive end
M.toggleRandomTornadoes = function() 
							doRandomTornadoes = not doRandomTornadoes 
							if doRandomTornadoes then 
								log("I", "", "Random tornadoes turned on.") 
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Random tornado spawning turned on.", config = {timeOut = 5000}})
							else
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Random tornado spawning turned off.", config = {timeOut = 5000}})
								log("I", "", "Random tornadoes turned off.")
							end
						  end
						  
M.toggleRandomPath = toggleRandomPath
M.setTornadoMovementSpeed = setTornadoMovementSpeed
M.toggleSetParams = function() 
						isSetStrength = not isSetStrength 
						if isSetStrength then
							guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Tornado using set parameters", config = {timeOut = 5000}})
						else
							guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Tornado using random parameters", config = {timeOut = 5000}})
						end
					  end
M.setParameters = setParameters


return M



